/*
    AQTRONIX C++ Library
    Copyright 2005-2006 Parcifal Aertssen

    This file is part of AQTRONIX C++ Library.

    AQTRONIX C++ Library is free software; you can redistribute it
	and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

    AQTRONIX C++ Library is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AQTRONIX C++ Library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
    MA  02111-1307  USA
*/
// HTTPFirewall.cpp: implementation of the CHTTPFirewall class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "HTTPFirewall.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHTTPFirewall::CHTTPFirewall(CHTTPFirewallSettings& _Settings, CLogger& _logger): Settings(_Settings),CFirewall(_Settings,_logger)
{

}

CHTTPFirewall::~CHTTPFirewall()
{

}

bool CHTTPFirewall::ScanRawHeaders(CString &Headers)
{

#define CR '\r'
#define LF '\n'

	bool hack(false);
	CString strhLog = Headers.Left(4096); //for log file

	//HEADERS: generic header length
	if(Settings.Headers.MaxLength>0){
		if(ScanVariableLength(Headers,Settings.Headers.MaxLength,CR,LF)){
			hack = true;
			Alert("Maximum generic header length exceeded");
		}
	}
		
	//HEADERS: check for encoding abuse
	if(Settings.Headers.DenyEncodingExploits){
		if(ScanHexEncodingAbuse(Headers)){
			hack = true;
			Alert("Encoding exploit in headers (abuse of: '%00' or '%u0000')");
		}
	}
	
	Headers = CURL::URLEncodedToASCII(Headers);

	//HEADERS: Deny encoding exploits (normalize twice, except for '+', and reject if change occurs)
	if(Settings.Headers.DenyEncodingExploits){
		if(CanHexDecode(Headers,false)){
			hack = true;
			Alert("Embedded encoding exploit in headers");
		}
	}

	//HEADERS: Deny directory traversal (parent path)
	if(Settings.Headers.DenyDirectoryTraversal){
		if(ScanDirectoryTraversal(Headers)){
			hack = true;
			Alert("Directory traversal not allowed in headers");
		}
	}

	//HEADERS: Deny high bit characters (shellcode)
	if(Settings.Headers.DenyHighBitShellcode){
		if(ScanHighBitShellCode(Headers)){
			hack = true;
			Alert("High bit characters (shellcode) not allowed in headers");
		}
	}

	CStringHelper::Safe_MakeLower(Headers);

	//HEADERS: Deny SQL injection (words in list should be lowercase)
	if(Settings.Headers.DenySQLInjection){	
		int sqlcount(Settings.SQLKeyWords.Find(";")!=NULL?-1:0);
		CStringList Matches;
		sqlcount += ScanSQLInjection(Headers,Settings.SQLKeyWords,Matches);
		if(sqlcount>0){
			if(sqlcount>1){
				hack = true;
				Alert("Possible SQL injection in headers: " + CStringHelper::ToString(Matches,","));
			}else{
				logger.Append("WARNING: SQL keyword found in headers (1 more will block request): " + CStringHelper::ToString(Matches,","));
			}
		}
	}

	//HEADERS: Deny certain sequences (words in list should be lowercase)
	if(Settings.Headers.DenySequences.Enabled){
		hack |= DenySequences(Headers,Settings.Headers.DenySequences.List,"'","' not allowed in headers");
	}

	//HEADERS: Deny regex
	if(Settings.Headers.DenyRegex.Enabled){
		hack |= DenyRegex(Headers,Settings.Headers.DenyRegex.Map,"'","' regex not allowed in headers");
	}

	if (hack){
		strhLog.Replace('\r',' ');
		strhLog.Replace('\n',' ');
		logger.Append(strhLog);
	}

	return hack;
}

bool CHTTPFirewall::ScanRawData(CString &Data)
{
	bool hack(false);
	CString strdLog = Data.Left(1024); //for log file

	//DATA: variable size
	if(Settings.Global.MaxPostVariableLength>0 && Data.Find('=')!=-1 && Data.Find('&')!=-1){ //TODO: ScanRawData - scan only urlencoded data
		if(ScanVariableLength(Data,Settings.Global.MaxPostVariableLength,'&','&')){
			hack = true;
			Alert("Maximum variable length exceeded in data");
		}
	}

	//DATA: check for encoding abuse
	if(Settings.Global.DenyPostEncodingExploits){
		if(ScanHexEncodingAbuse(Data)){
			hack = true;
			Alert("Encoding exploit in data (abuse of: '%00' or '%u0000')");
		}
	}

	Data = CURL::URLEncodedToASCII(Data);

	//DATA: Deny encoding exploits (normalize twice, except for '+', and reject if change occurs)
	if(Settings.Global.DenyPostEncodingExploits){
		if(CanHexDecode(Data,false)){
			hack = true;
			Alert("Encoding exploit in data (embedded encoding)");
		}		
	}

	//DATA: Deny directory traversal (parent path)
	if(Settings.Global.DenyPostDirectoryTraversal){
		if(ScanDirectoryTraversal(Data)){
			hack = true;
			Alert("Directory traversal not allowed in data");
		}
	}

	//DATA: Deny high bit characters (shellcode)
	if(Settings.Global.DenyPostHighBitShellcode){
		if(ScanHighBitShellCode(Data)){
			hack = true;
			Alert("High bit characters (shellcode) not allowed in data");
		}
	}

	CStringHelper::Safe_MakeLower(Data);

	//DATA: Deny SQL injection (words in list should be lowercase)
	if(Settings.Global.DenyPostSQLInjection){
		int sqlcount(0);
		CStringList Matches;
		sqlcount += ScanSQLInjection(Data,Settings.SQLKeyWords, Matches);
		if(sqlcount>0){
			if(sqlcount>1){
				hack = true;
				Alert("Possible SQL injection in data: " + CStringHelper::ToString(Matches,","));
			}else{
				logger.Append("WARNING: SQL keyword found in data (1 more will block request): " + CStringHelper::ToString(Matches,","));
			}
		}
	}

	//DATA: deny certain sequences (list should be lowercase)
	if(Settings.Global.DenyPostSequences.Enabled){
		hack |= DenySequences(Data,Settings.Global.DenyPostSequences.List,"'","' not allowed in data");
	}

	//DATA: deny regex
	if(Settings.Global.DenyPostRegex.Enabled){
		hack |= DenyRegex(Data,Settings.Global.DenyPostRegex.Map,"'","' regex not allowed in data");
	}

	if(hack){
		strdLog.Replace('\r',' ');
		strdLog.Replace('\n',' ');
		logger.Append(strdLog);
	}

	return hack;
}

bool CHTTPFirewall::ScanAccount(CString &User)
{
	if(User==""){
		return false;
	}else{
		CString USER(User);
		CStringHelper::Safe_MakeUpper(USER);
		bool hack(false);

		//AUTH - block logon attempts for certain critical accounts
		if(Settings.Authentication.DenySystemAccounts){
			CString UserLeft = USER.Left(5);
			if(	UserLeft=="IUSR_" ||	//InternetUser
				UserLeft=="IWAM_" ||	//InternetWAM
				UserLeft=="VUSR_" ||	//Visual studio server components
				USER=="SYSTEM" ||		//SYSTEM account
				USER=="KRBTGT" ||		//Kerberos
				USER=="ASPNET" ||		//ASP.NET (out of process)
				USER=="ASPNETUSER" ||	//ASP.NET (out of process) Windows 2003
				USER=="TSINTERNETUSER" ||//Terminal services account (internet account)
				USER=="NETWORK SERVICE"){
					hack = true;
					Alert("Attempt to authenticate with system critical account");
			}
		}

		//AUTH - only allow certain accounts
		if(Settings.Authentication.AllowAccounts.Enabled){
			if(!IsInListCompareNoCase(User,Settings.Authentication.AllowAccounts.List)){
				hack = true;
				Alert("Account not in allowed list");
			}
		}

		//AUTH - Deny certain accounts
		if(Settings.Authentication.DenyAccounts.Enabled){
			if(IsInListCompareNoCase(User,Settings.Authentication.DenyAccounts.List)){
				hack = true;
				Alert("Account not allowed to authenticate");
			}
		}
		return hack;
	}
}

bool CHTTPFirewall::ScanFilename(CFileName &fn)
{
	bool hack(false);
	char c;
	CString str;

	//FILENAME: Deny default document
	if(Settings.Filename.DenyDefaultDocument){
		if(fn.FileName == ""){
			hack = true;
			Alert("Default document not allowed");
		}
	}

	//FILENAME: Deny certain characters in the filename (characters are case sensitive!)
	for(int i=0;i<Settings.Filename.DenyCharacters.GetLength();i++){
		c = Settings.Filename.DenyCharacters.GetAt(i);
		hack |= DenyChar(fn.FileName,c,"'" + CString(c) + "' not allowed in filename");
	}

	//FILENAME: Deny certain cgi applications (applications in list should be lowercase).
	if(Settings.Filename.DenyFiles.Enabled){
		hack |= DenySequences(fn.FileNameLCase,Settings.Filename.DenyFiles.List,"accessing/running '","' file");
	}

	//EXTENSION: Only allow certain extensions (extensions in the list should be lowercase)
	if(Settings.Extensions.AllowExtensions.Enabled){
		if(!IsInList(fn.ExtensionLCase,Settings.Extensions.AllowExtensions.List)){
			hack = true;
			Alert("accessing/running '" + fn.Extension + "' file");
		}
	}

	//EXTENSION: Deny certain extensions (extensions in the list should be lowercase)
	if(Settings.Extensions.DenyExtensions.Enabled){
		hack |= DenySequences(fn.ExtensionLCase,Settings.Extensions.DenyExtensions.List,"accessing/running '","' file");
	}

	return hack;
}

bool CHTTPFirewall::ScanURL(const char *p)
{
	bool hack(false);
	CString URL(p);
	CStringHelper::Safe_MakeLower(URL);

	//URL: Only allow certain characters as the start of a URL (list should be lowercase)
	if(Settings.URL.AllowURL.Enabled){
		if(!IsStartInList(URL,Settings.URL.AllowURL.List)){
			hack = true;
			Alert("URL not in allowed list");
		}
	}

	//URL: Deny parent path
	if (Settings.URL.DenyParentPath)
		hack |= DenyString(p,"..","'..' not allowed in URL (parent path)");

	//URL: Deny trailing dot in directory
	if (Settings.URL.DenyTrailingDotInDir)
		hack |= DenyString(p,"./","'./' not allowed in URL (trailing dot on a directory name)");
	
	//URL: Deny backslash
	if (Settings.URL.DenyBackslash)
		hack |= DenyChar(p,'\\',"'\\' not allowed in URL (backslashes)");

	//URL: Deny alternate stream access
	if (Settings.URL.DenyAlternateStream)
		hack |= DenyChar(p,':',"':' not allowed in URL (alternate stream access)");

	//URL: Deny escaping
	if (Settings.URL.DenyEscaping)
		hack |= DenyChar(p,'%',"'%' not allowed in URL (hex encoding)");

	//URL: Deny running multiple applications
	if (Settings.URL.DenyMultipleCGI)
		hack |= DenyChar(p,'&',"'&' not allowed in URL (running multiple CGI processes in single request)");
	
	//URL: Deny certain characters
	char c;	int i;
	for(i=0;i<Settings.URL.DenyCharacters.GetLength();i++){
		c = Settings.URL.DenyCharacters.GetAt(i);
		hack |= DenyChar(p,c,"'" + CString(c) + "' not allowed in URL");
	}

	//URL: Deny high bit characters (shellcode)
	if(Settings.URL.DenyHighBitShellcode){
		if(ScanHighBitShellCode(p)){
			hack = true;
			Alert("High bit characters (shellcode) not allowed in URL");
		}
	}

	//URL: Deny special whitespace characters
	if(Settings.URL.DenySpecialWhitespace){
		hack |= DenySpecialWhitespace(p, " not allowed in URL");
	}

	//URL: Deny certain sequences (list should be lowercase)
	if(Settings.URL.DenySequences.Enabled){
		hack |= DenySequences(URL,Settings.URL.DenySequences.List,"'","' not allowed in URL");
	}

	//URL: deny regex
	if(Settings.URL.DenyRegex.Enabled){
		hack |= DenyRegex(URL,Settings.URL.DenyRegex.Map,"'","' regex not allowed in URL");
	}
	
	return hack;
}

bool CHTTPFirewall::ScanRawQuerystring(CString& QueryString)
{
	int l(0);
	if((l = QueryString.GetLength())==0){
		return false;
	}else{

		bool hack(false);
		logger.Append(QueryString.Left(9000));

		//QUERYSTRING: size
		if(Settings.Limits.LimitQueryString){
			if (l>Settings.Limits.MaxQueryString){
				hack = true;
				Alert("QueryString is too long!");
			}
		}

		//QUERYSTRING: variable size
		if(Settings.QueryString.MaxVariableLength>0 && QueryString.Find('=')!=-1){
			if(ScanVariableLength(QueryString,Settings.QueryString.MaxVariableLength,'&','&')){
				hack = true;
				Alert("Maximum variable length exceeded in querystring");
			}
		}

		//QUERYSTRING: encoding abuse
		if(Settings.QueryString.DenyEncodingExploits){
			if(ScanHexEncodingAbuse(QueryString)){
				hack = true;
				Alert("Encoding exploit in querystring (abuse of: '%00' or '%u0000')");
			}
		}
		
		QueryString = CURL::URLEncodedToASCII(QueryString);

		//QUERYSTRING: Encoding Exploits (normalize twice, except for '+', and reject if change occurs)
		if(Settings.QueryString.DenyEncodingExploits){
			if(CanHexDecode(QueryString,false)){
				hack = true;
				Alert("Querystring embedded encoding exploit");
			}
		}

		//QUERYSTRING: Deny directory traversal (parent path)
		if(Settings.QueryString.DenyDirectoryTraversal){
			if(ScanDirectoryTraversal(QueryString)){
				hack = true;
				Alert("Directory traversal not allowed in querystring");
			}
		}

		//QUERYSTRING: Deny high bit characters (shellcode)
		if(Settings.QueryString.DenyHighBitShellcode){
			if(ScanHighBitShellCode(QueryString)){
				hack = true;
				Alert("High bit characters (shellcode) not allowed in querystring");
			}
		}

		//QUERYSTRING: Deny special whitespace characters
		if(Settings.QueryString.DenySpecialWhitespace){
			hack |= DenySpecialWhitespace(QueryString," not allowed in querystring");
		}

		CStringHelper::Safe_MakeLower(QueryString);
		
		//QUERYSTRING: SQL injection (words in list should be lowercase)
		if(Settings.QueryString.DenySQLInjection){
			int sqlcount(0);
			CStringList Matches;
			sqlcount += ScanSQLInjection(QueryString,Settings.SQLKeyWords,Matches);
			if(sqlcount>0){
				if(sqlcount>1){
					hack = true;
					Alert("Possible SQL injection in querystring: " + CStringHelper::ToString(Matches,","));
				}else{
					logger.Append("WARNING: SQL keyword found in querystring (1 more will block request): " + CStringHelper::ToString(Matches,","));
				}
			}
		}

		//QUERYSTRING: deny certain sequences (list should be lowercase)
		if(Settings.QueryString.DenySequences.Enabled){
			hack |= DenySequences(QueryString,Settings.QueryString.DenySequences.List,"'","' not allowed in querystring");
		}

		//QUERYSTRING: deny regex
		if(Settings.QueryString.DenyRegex.Enabled){
			hack |= DenyRegex(QueryString,Settings.QueryString.DenyRegex.Map,"'","' regex not allowed in data");
		}

		return hack;
	}
}

bool CHTTPFirewall::ScanConnection(CString &IP, bool& is_blacklisted, bool& is_monitored)
{
	bool hack(false);

	//CONNECTION - IP blocking
	if(Settings.Connection.DenyAddresses.Enabled){
		if(Settings.DenyIPRangeList.IsInList(IP)){
			hack = true;
			is_blacklisted = true;
			Alert("IP address '" + IP + "' not allowed");
		}
	}

	//CONNECTION - Request Limits
	if(Settings.Connection.RequestsLimit.Enabled){
		//add request to cache
		IP_RequestsLimit.Add(IP);

		//block if number of requests exceeds certain treshold
		if(IP_RequestsLimit.Count(IP)>Settings.Connection.RequestsLimit.MaxCount){
			hack = true;
			Alert("Reached connection request limit");
		}
	}

	//CONNECTION - Hacking IP Blocking
	if(Settings.HTTPResponse.BlockIP.Enabled){
		if(IP_HACK_Blocked.Count(IP)>Settings.HTTPResponse.BlockIP.MaxCount){
			hack = true;
			is_blacklisted = true;
			Alert("Blocked IP address (previous alert)");
		}
	}

	//RESPONSE MONITOR
	if(Settings.ResponseMonitor.BlockClientErrors.Enabled){
		if(IP_RESPONSE_Client_Errors.Count(IP)>Settings.ResponseMonitor.BlockClientErrors.MaxCount){
			hack = true;
			is_blacklisted = true;
			Alert("Blocked IP address (multiple HTTP client errors)");
		}
	}
	if(Settings.ResponseMonitor.BlockServerErrors.Enabled){
		if(IP_RESPONSE_Server_Errors.Count(IP)>Settings.ResponseMonitor.BlockServerErrors.MaxCount){
			hack = true;
			is_blacklisted = true;
			Alert("Blocked IP address (multiple HTTP server errors)");
		}
	}

	//CONNECTION - Hacking IP Monitoring
	if(Settings.HTTPResponse.MonitorIP){
		if(IP_HACK_Monitored.Exists(IP)){
			is_monitored = true;
			logger.Append("MONITORED: IP address (previous alert)");
		}
	}

	//CONNECTION - IP Monitoring
	if(Settings.Connection.MonitorAddresses.Enabled){
		if(Settings.MonitorIPRangeList.IsInList(IP)){
			is_monitored = true;
			logger.Append("MONITORED: IP address '" + IP + "'");
		}
	}

	return hack;
}

bool CHTTPFirewall::ScanAuthentication(CString &User, CString &Pass, CString& IP)
{
	bool hack(false);

	//AUTH - block blank password
	if(Settings.Authentication.DenyBlankPasswords && Pass==""){
		hack = true;
		Alert("Blank passwords not allowed");
	}
	//AUTH - block password=username
	if(Settings.Authentication.DenySamePasswordAsUsername && Pass.CompareNoCase(User)==0){
		hack = true;
		Alert("Password is same as username");
	}
	//AUTH - block default password lists
	if(Settings.Authentication.DenyDefaultPasswords.Enabled){
		if(Settings.Authentication.DenyDefaultPasswords.List.Find(Pass)!=NULL){
			hack = true;
			Alert("Default password '" + Pass + "' not allowed");
		}
	}
	
	//AUTH - Scan Account
	hack |= ScanAccount(User);

	CStringHelper::Safe_MakeUpper(User);
		
	//AUTH - scan for DoS on authentication of accounts (lookup)
	if(Settings.Authentication.DenyBruteForceAttack.Enabled){
		CString ipuser; ipuser=IP+User;
		//add this ip to cache
		if(IP_Authentication.Count(ipuser,Pass)==0){
			IP_Authentication.Add(ipuser,Pass);
		}
		//count ip references in cache
		if(IP_Authentication.Count(ipuser)>Settings.Authentication.DenyBruteForceAttack.MaxCount){
			hack = true;
			Alert("Possible brute force attack or DoS.");
		}
	}

	return hack;
}

bool CHTTPFirewall::ScanVerb(CString Verb)
{
	bool hack(false);
	logger.Append(Verb.Left(2048));

	//VERB: only allow certain verbs (verbs in list are case sensitive!)
	if(Settings.Verbs.AllowVerbs.Enabled){
		if(!IsInList(Verb,Settings.Verbs.AllowVerbs.List)){
			hack = true;
			Alert("HTTP VERB not allowed");
		}
	}
	
	//VERB: Deny certain verbs (verbs in list should be uppercase)
	CStringHelper::Safe_MakeUpper(Verb); //Also deny lowercase & variants!
	if(Settings.Verbs.DenyVerbs.Enabled){
		if(IsInList(Verb,Settings.Verbs.DenyVerbs.List)){
			hack = true;
			Alert("HTTP VERB not allowed");
		}
	}

	return hack;
}

bool CHTTPFirewall::ScanRawURL(CURL &RawUrl)
{
	bool hack(false);
	CString URLDecoded = CURL::URLEncodedToASCII(RawUrl.Url);

	//URL: RFC compliant
	if(Settings.URL.RFCCompliantURL){
		if(!RawUrl.IsRFCCompliant()){
			hack = true;
			logger.Append(RawUrl.RawUrl.Left(4096));
			Alert("URL is not RFC compliant!");
		}
	}

	//URL: HTTP RFC compliant
	if(Settings.URL.RFCCompliantHTTPURL){
		if(!RawUrl.IsHTTPCompliantUrl()){
			hack = true;
			logger.Append(RawUrl.RawUrl.Left(4096));
			Alert("URL is not not RFC compliant (HTTP URL cannot contain authentication)!");
		}
	}

	//URL
	logger.Append(RawUrl.Url.Left(4096));

	//URL Scanning (RAW)
	if(Settings.URL.UseRawScan){

		//URL: Check size
		if(Settings.Limits.LimitURL){
			if (RawUrl.Url.GetLength() > Settings.Limits.MaxURL){
				hack = true;
				Alert("URL is too long!");
			}
		}

		//URL: check for encoding abuse
		if(Settings.URL.DenyEncodingExploits){
			if(ScanHexEncodingAbuse(RawUrl.Url)){
				hack = true;
				Alert("Encoding exploit in URL (abuse of: '%00' or '%u0000')");
			}
		}

		//URL: Encoding Exploits (normalize twice, except for '+', and reject if change occurs)
		if(Settings.URL.DenyEncodingExploits){
			if(CanHexDecode(URLDecoded,false)){
				hack = true;
				Alert("URL embedded encoding exploit");
			}
		}

		//URL: scan
		hack |= ScanURL(URLDecoded);
	}

	//FILENAME (RAW)
	if(Settings.Filename.UseRawScan){
		hack |= ScanFilename(CFileName(URLDecoded));
	}

	//QUERYSTRING (RAW)
	if(Settings.QueryString.UseRawScan && RawUrl.Querystring!=""){
		hack |= ScanRawQuerystring(RawUrl.Querystring);
	}

	return hack;
}

bool CHTTPFirewall::ScanVersion(const CString &Version, const CString& Host)
{
	bool hack(false);
	logger.Append(Version.Left(1024));

	//VERSION: Check size
	if(Settings.Limits.LimitHTTPVersion){
		if (Version.GetLength() > Settings.Limits.MaxHTTPVersion){
			hack = true;
			Alert("HTTP version too long");
		}
	}
		
	//VERSION: allowed HTTP versions
	if(Settings.Limits.AllowHTTPVersions.Enabled){
		if(!IsInList(Version,Settings.Limits.AllowHTTPVersions.List)){
			hack = true;
			Alert("HTTP version not in allowed list");
		}
	}

	//TODO: denied http versions?
	
	//VERSION: host header required for HTTP 1.1 (most likely worms/script kiddies/other stuff that is not RFC compliant)
	if(Settings.Headers.RFCCompliantHostHeader){
		if(Version=="HTTP/1.1"){
			if(Host==""){
				hack = true;
				Alert("'Host:' header required for HTTP 1.1 request");
			}
			/*
		}else{
			if((Version=="HTTP/0.9" || Version=="") && Host!=""){
				hack=true;
				Alert("'Host:' header not allowed in HTTP 0.9");
			}//*/
		}
	}

	return hack;
}

bool CHTTPFirewall::ScanCookie(CString &Cookie)
{
	if(Cookie.GetLength()==0){
		return false;
	}else{

		bool hack(false);
		logger.Append(Cookie);

		//COOKIE: variable size
		if(Settings.Cookie.MaxVariableLength>0){
			if(ScanVariableLength(Cookie,Settings.Cookie.MaxVariableLength,' ',' ')){
				hack = true;
				Alert("Maximum variable length exceeded in cookie");
			}
		}
	
		//COOKIE: check for encoding abuse
		if(Settings.Cookie.DenyEncodingExploits){
			if(ScanHexEncodingAbuse(Cookie)){
				hack = true;
				Alert("Encoding exploit in cookie (abuse of: '%00' or '%u0000')");
			}
		}

		Cookie = CURL::URLEncodedToASCII(Cookie); //can be encoded: http://wp.netscape.com/newsref/std/cookie_spec.html

		//COOKIE: Deny encoding exploits (normalize twice, except for '+', and reject if change occurs)
		if(Settings.Cookie.DenyEncodingExploits){
			if(CanHexDecode(Cookie,false)){
				hack = true;
				Alert("Encoding exploit in cookie (embedded encoding)");
			}
		}

		//COOKIE: Deny directory traversal (parent path)
		if(Settings.Cookie.DenyDirectoryTraversal){
			if(ScanDirectoryTraversal(Cookie)){
				hack = true;
				Alert("Directory traversal not allowed in cookie");
			}
		}

		//COOKIE: Deny high bit characters (shellcode)
		if(Settings.Cookie.DenyHighBitShellcode){
			if(ScanHighBitShellCode(Cookie)){
				hack = true;
				Alert("High bit characters (shellcode) not allowed in cookie");
			}
		}

		//COOKIE: Deny special whitespace characters
		if(Settings.Cookie.DenySpecialWhitespace){
			hack |= DenySpecialWhitespace(Cookie," not allowed in cookie");
		}

		CStringHelper::Safe_MakeLower(Cookie);

		//COOKIE: Deny SQL injection
		if(Settings.Cookie.DenySQLInjection){
			int sqlcount( Settings.SQLKeyWords.Find(";")!=NULL?-1:0 );
			CStringList Matches;
			sqlcount += ScanSQLInjection(Cookie,Settings.SQLKeyWords,Matches);
			if(sqlcount>0){
				if(sqlcount>1){
					hack = true;
					Alert("Possible SQL injection in cookie: " + CStringHelper::ToString(Matches,","));
				}else{
					logger.Append("WARNING: SQL keyword found in cookie (1 more will block request): " + CStringHelper::ToString(Matches,","));
				}
			}			
		}

		//COOKIE: deny certain sequences (list should be lowercase)
		if(Settings.Cookie.DenySequences.Enabled){
			hack |= DenySequences(Cookie,Settings.Cookie.DenySequences.List,"'","' not allowed in cookie");
		}

		return hack;
	}
}

bool CHTTPFirewall::ScanPath(CFileName& fn)
{
	bool hack(false);
	logger.Append(fn.FullPath.Left(9000));

	//FULLPATH
	if(Settings.Path.DenyParentPath)
		hack |= DenyString(fn.FullPath,"..","Parent Path attempt ('..')");

	if(Settings.Path.DenyEscaping)
		hack |= DenyChar(fn.FullPath,'%',"Possible encoding exploit");

	if(Settings.Path.DenySpecialWhitespace)
		hack |= DenySpecialWhitespace(fn.FullPath," not allowed in path");

	if(Settings.Path.DenyMultipleColons){
		hack |= DenyMultipleCharacters(fn.FullPath,':',"Multiple colons not allowed in path");
	}

	int i; char c; CString str;
	
	for(i=0;i<Settings.Path.DenyCharacters.GetLength();i++){
		c = Settings.Path.DenyCharacters.GetAt(i);
		hack |= DenyChar(fn.FullPath,c,"'" + CString(c) + "' not allowed in path");
	}

	//PATH: Deny dot in path (!= full path)
	if(Settings.Path.DenyDot)
		hack |= DenyChar(fn.Path,'.',"'.' not allowed in path");
	
	//PATH: Only allow certain paths (I use full paths here, because when
	//      using path, virtual dirs don't work if they are set to c:\somedir
	//      instead of c:\somedir\											*/
	if(Settings.Path.AllowPaths.Enabled){ //(lowercase!)
		if(!IsStartInList(fn.FullPathLCase,Settings.Path.AllowPaths.List)){
			hack = true;
			Alert("Not in allowed path list '" + fn.FullPath + "'");
		}
	}

	//FILENAME
	hack |= ScanFilename(fn);

	return hack;
}

bool CHTTPFirewall::ScanUserAgent(CString &UserAgent, CString& IP)
{
	bool hack(false);

	//User-Agent 		=	"User-Agent" ":" 1*( product | comment )
	//product 			=	token ["/" product-version ]
	//product-version 	=	token
	//comment 			=	"(" *( ctext | comment ) ")"
	//ctext 			=	<any TEXT excluding "(" and ")">
	//token 			=	1*<any CHAR except CTLs or tspecials>
	//tspecials 		=	"(" | ")" | "<" | ">" | "@" | "," | ";" | ":" | "\" | <"> | "/" | "[" | "]" | "?" | "=" | "{" | "}" | SP | HT

	//multiple User-Agents (or any headers exception):
	//Multiple message-header fields with the same field-name MAY be present in a message
	//if and only if the entire field-value for that header field is defined as a comma-separated
	//list [i.e., #(values)]. It MUST be possible to combine the multiple header fields into
	//one "field-name: field-value" pair, without changing the semantics of the message, by
	//appending each subsequent field-value to the first, each separated by a comma.

	if(UserAgent.GetLength()==0){
		//UA: Deny empty
		if(Settings.UserAgent.DenyEmpty){
			hack = true;
			Alert("Empty User Agent not allowed");
		}
	}else{

		//UA: require certain characters ('/ -.')
		if(Settings.UserAgent.RequireCharacter.GetLength()>0){
			if(UserAgent.FindOneOf(Settings.UserAgent.RequireCharacter)==-1){
				hack = true;
				Alert("User Agent requires at least one of these characters: '" + Settings.UserAgent.RequireCharacter + "'");
			}
		}

		//UA: Deny User Agent not RFC compliant
		if(Settings.UserAgent.DenyNonRFC){
			//remove comments
			CString UserAgentClean = RemoveRFCComments(UserAgent);
			//logger.Append(UserAgentClean);
			//scan for non-token chars
			//if(UserAgentClean.FindOneOf("()<>@,;:\\\"/[]?={} \t")!=-1){
			if(UserAgentClean.FindOneOf("()<>@;:\\\"[]?={}\t")!=-1){	//allow '/' + ' ' + ',' (multiple headers with same name)
				hack = true;
				Alert("User Agent not RFC compliant");
			}
		}

		//UA: Deny high bit characters (shellcode)
		if(Settings.UserAgent.DenyHighBitShellcode){
			if(ScanHighBitShellCode(UserAgent)){
				hack = true;
				Alert("High bit characters (shellcode) not allowed in User Agent");
			}
		}

		//UA: Deny special whitespace characters
		if(Settings.UserAgent.DenySpecialWhitespace){
			hack |= DenySpecialWhitespace(UserAgent," not allowed in User Agent");
		}

		//UA: Deny switching User-Agent
		if(Settings.UserAgent.BlockSwitching.Enabled){
			//add to cache
			IP_UserAgent.AddNotInCache(IP,UserAgent);

			//block if number of requests exceeds certain treshold
			if(IP_UserAgent.Count(IP)>Settings.UserAgent.BlockSwitching.MaxCount){
				hack = true;
				Alert("User Agent is switching too much on this IP address");
			}
		}

		//UA: Denied User-Agents (case insensitive)
		if(Settings.UserAgent.DenyUserAgents.Enabled){
			if(IsInListCompareNoCase(UserAgent,Settings.UserAgent.DenyUserAgents.List)){
				hack = true;
				Alert("User Agent not allowed");
			}
		}

		//UA: Denied User-Agent Sequences (list should be lowercase)
		if(Settings.UserAgent.DenySequences.Enabled){
			CStringHelper::Safe_MakeLower(UserAgent);
			hack |= DenySequences(UserAgent,Settings.UserAgent.DenySequences.List,"'","' not allowed in User Agent");
		}
	}

	return hack;
}

bool CHTTPFirewall::ScanContent(CString &ContentType, CString &ContentLength)
{
	bool hack(false);
	int l(0);

	//HEADERS: Check Content-Length value
	if((l = ContentLength.GetLength())!=0 && Settings.Limits.LimitContentLength){
		//to prevent overflow or other types of attack, block request if the length
		//of Content-Length header is longer than length of max allowed value
		//or longer than 9 characters (maximum value of long integer - 1 character)!
		if(l > Settings.Limits.MaxContentLength.GetLength() || l > 9){
			hack = true;
			Alert("Content-Length is too long!");
			logger.Append(ContentLength.Left(1024));
		}else{
			if (atol(ContentLength) > atol(Settings.Limits.MaxContentLength)){
				hack = true;
				Alert("Content-Length is too long!");
				logger.Append(ContentLength);
			}
		}
	}

	//HEADERS: Allowed Content-Types
	if(Settings.ContentType.AllowContentTypes.Enabled){
		if(!IsNotEmptyStartInList(ContentType,Settings.ContentType.AllowContentTypes.List)){
			hack = true;
			Alert("Content-Type '" + ContentType.Left(2048) + "' not allowed");
		}
	}

	//HEADERS: Denied Content-Types
	if(Settings.ContentType.DenyContentTypes.Enabled && ContentType.GetLength()>0){
		if(IsStartInList(ContentType,Settings.ContentType.DenyContentTypes.List)){
			hack = true;
			Alert("Content-Type '" + ContentType.Left(2048) + "' not allowed");
		}
	}
	
	return hack;
}

bool CHTTPFirewall::IsExcludedSite(CString &instance)
{
	if(Settings.Engine.ExcludedInstances.Enabled && IsInList(instance,Settings.Engine.ExcludedInstances.List)){
		logger.Append("Web site instance is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedHost(CString& Host)
{
	if(Settings.Engine.ExcludedHosts.Enabled && IsInList(Host,Settings.Engine.ExcludedHosts.List)){
		logger.Append("Web site host is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedIP(CString& IP)
{
	if(Settings.Engine.ExcludedIPAddresses.Enabled && Settings.ExcludedIPRangeList.IsInList(IP)){
		logger.Append("IP address is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedURL(CString& url, CString& Querystring)
{
	if	(
			(Settings.URL.ExcludedURL.Enabled && IsStartInList(url,Settings.URL.ExcludedURL.List)) ||
			(Settings.QueryString.ExcludedQueryString.Enabled && IsInList(Querystring,Settings.QueryString.ExcludedQueryString.List))
		)
	{
		logger.Append(CString(url + '?' + Querystring).Left(2048));
		logger.Append("URL is excluded from scanning");
		return true;
	}
	return false;
}

void CHTTPFirewall::LoadCache()
{
	CFirewall::LoadCache();

	//(re-)load IP Lists
	Settings.MonitorIPRangeList.Enumerate(Settings.Connection.MonitorAddresses.List);
	Settings.DenyIPRangeList.Enumerate(Settings.Connection.DenyAddresses.List);
	Settings.ExcludedIPRangeList.Enumerate(Settings.Engine.ExcludedIPAddresses.List);

	//stringcaches
	IP_Authentication.SetMaxAge(CTimeSpan(0,0,Settings.Authentication.DenyBruteForceAttack.MaxTime>1?Settings.Authentication.DenyBruteForceAttack.MaxTime:1,0));//in min.
	IP_RequestsLimit.SetMaxAge(CTimeSpan(0,0,Settings.Connection.RequestsLimit.MaxTime>1?Settings.Connection.RequestsLimit.MaxTime:1,0));	//in min.

	IP_HACK_Monitored.SetMaxAge(CTimeSpan(0,Settings.HTTPResponse.MonitorIPTimeout>1?Settings.HTTPResponse.MonitorIPTimeout:1,0,0));//in hours.
	IP_HACK_Blocked.SetMaxAge(CTimeSpan(0,Settings.HTTPResponse.BlockIP.MaxTime>1?Settings.HTTPResponse.BlockIP.MaxTime:1,0,0));//in hours.

	Extension_RequestsLimit.SetMaxAge(CTimeSpan(0,0,Settings.Extensions.RequestsLimit.MaxTime>1?Settings.Extensions.RequestsLimit.MaxTime:1,0));	//in min.
	IP_UserAgent.SetMaxAge(CTimeSpan(0,0,Settings.UserAgent.BlockSwitching.MaxTime>1?Settings.UserAgent.BlockSwitching.MaxTime:1,0));	//in min.

	IP_RESPONSE_Client_Errors.SetMaxAge(CTimeSpan(0,0,Settings.ResponseMonitor.BlockClientErrors.MaxTime>1?Settings.ResponseMonitor.BlockClientErrors.MaxTime:1,0));	//in min.
	IP_RESPONSE_Server_Errors.SetMaxAge(CTimeSpan(0,0,Settings.ResponseMonitor.BlockServerErrors.MaxTime>1?Settings.ResponseMonitor.BlockServerErrors.MaxTime:1,0));	//in min.
}

void CHTTPFirewall::ClearCache()
{
	CFirewall::ClearCache();

	//(re-)load IP Lists
	Settings.MonitorIPRangeList.Enumerate(Settings.Connection.MonitorAddresses.List);
	Settings.DenyIPRangeList.Enumerate(Settings.Connection.DenyAddresses.List);
	Settings.ExcludedIPRangeList.Enumerate(Settings.Engine.ExcludedIPAddresses.List);

	//stringcaches
	IP_Authentication.Clear();
	IP_RequestsLimit.Clear();

	IP_HACK_Monitored.Clear();
	IP_HACK_Blocked.Clear();

	Extension_RequestsLimit.Clear();
	IP_UserAgent.Clear();

	IP_RESPONSE_Client_Errors.Clear();
	IP_RESPONSE_Server_Errors.Clear();
}

void CHTTPFirewall::SetLoggerFields()
{
	//set logger fields
	CStringList fields;
	fields.AddTail("Date");
	fields.AddTail("Time");
	fields.AddTail("Site Instance");
	fields.AddTail("Event");
	if(Settings.Logging.LogClientIP) fields.AddTail("Client IP");
	if(Settings.Logging.LogUserName) fields.AddTail("Username");
	if(Settings.HTTPLogging.Log_HTTP_VIA) fields.AddTail("Via header");
	if(Settings.HTTPLogging.Log_HTTP_X_FORWARDED_FOR) fields.AddTail("HTTP_X_FORWARDED_FOR");
	if(Settings.HTTPLogging.LogHostHeader) fields.AddTail("Host header");
	if(Settings.HTTPLogging.LogUserAgent) fields.AddTail("User agent");
	fields.AddTail("Additional info about request (event specific)");
	logger.SetHeaderFields(fields);
}

bool CHTTPFirewall::ScanRawRefererURL(LPCTSTR ReferrerUrl,LPCTSTR RequestedURLDecoded, CString& HostHeader)
{
	CURL RefUrl(ReferrerUrl);
	CFileName fn(RequestedURLDecoded);
	return ScanRawRefererURL(RefUrl,fn,HostHeader);
}

bool CHTTPFirewall::ScanRawRefererURL(CURL &ReferrerUrl, CFileName& RequestedFile, CString& HostHeader)
{
	bool hack(false);

	if(Settings.Referrer.UseScan){

		CString URLDecoded = CURL::URLEncodedToASCII(ReferrerUrl.Url);

		//REFERER URL
		logger.Append(ReferrerUrl.RawUrl.Left(4096));
		
		//REFERER URL: RFC compliant
		if(Settings.Referrer.RFCCompliantURL){
			if(!ReferrerUrl.IsRFCCompliant()){
				hack = true;
				Alert("Referrer URL is not RFC compliant!");
			}
		}

		//REFERER URL: HTTP RFC compliant
		if(Settings.Referrer.RFCCompliantHTTPURL){
			if(!ReferrerUrl.IsHTTPCompliantUrl()){
				hack = true;
				Alert("Referrer URL is not not RFC compliant (HTTP URL cannot contain authentication)!");
			}
		}

		//REFERER URL: check for encoding abuse & exploits
		if(Settings.Referrer.DenyEncodingExploits){
			//REFERER URL: check for encoding abuse
			if(ScanHexEncodingAbuse(ReferrerUrl.Url)){
				hack = true;
				Alert("Encoding exploit in Referer URL (abuse of: '%00' or '%u0000')");
			}

			//REFERER URL: Encoding Exploits (normalize twice, except for '+', and reject if change occurs)
			if(CanHexDecode(URLDecoded,false)){
				hack = true;
				Alert("Referrer URL embedded encoding exploit");
			}
		}

		//REFERER URL: scan
		hack |= ScanRefererURL(URLDecoded,RequestedFile,HostHeader);

		/*
		//FILENAME (RAW)
		if(Settings.UseFilenameRawScan){
			hack |= ScanFilename(CFileName(URLDecoded));
		}

		//QUERYSTRING (RAW)
		if(Settings.UseQuerystringRawScan && ReferrerUrl.Querystring!=""){
			hack |= ScanRawQuerystring(ReferrerUrl.Querystring);
		}//*/
	}

	return hack;
}

bool CHTTPFirewall::ScanRefererURL(CString& ReferrerURL, CFileName& RequestedFile, CString& HostHeader)
{
	bool hack(false);

	//REFERER URL: Deny deep linking
	if(Settings.Referrer.DenyHotLinking){
		//is certain url or file extension?
		if(IsListItemInString(RequestedFile.FullPathLCase,Settings.Referrer.HotLinkingUrls) || IsInList(RequestedFile.ExtensionLCase,Settings.Referrer.HotLinkingFileExtensions)){
			CString RefDomain = CURL::FQDNNoEndDot(ReferrerURL);
			CStringHelper::Safe_MakeLower(RefDomain);
			if(ReferrerURL.GetLength()>0){
				if(Settings.Referrer.HotLinkingUseHostHeader && RefDomain.CompareNoCase(HostHeader)==0 ){
					//allow: HostHeader == RefDomain
				}else{
					//REFERRER: No Host header
					if(Settings.Referrer.HotLinkingUseHostHeader && HostHeader.GetLength()==0){
						logger.Append("WARNING: Hot linking 'Host:' header missing!");
					}
					//REFERRER: Allowed domains (list should be lowercase)
					if(Settings.Referrer.HotLinkingAllowDomains.Enabled){
						if(!IsInList(RefDomain,Settings.Referrer.HotLinkingAllowDomains.List)){
							hack = true;
							Alert("Hot linking domain not in allowed list");
						}
					}
					//REFERRER: Denied domains (list should be lowercase)
					if(Settings.Referrer.HotLinkingDenyDomains.Enabled){
						if(IsInList(RefDomain,Settings.Referrer.HotLinkingDenyDomains.List)){
							hack = true;
							Alert("Hot linking domain in denied list");
						}
					}
				}
			}else{
				//REFERRER: Blank header
				if(Settings.Referrer.HotLinkingDenyBlankReferrer){
					//also blocks some proxy servers and browsers with certain security applications
					hack = true;
					Alert("Hot linking with blank referrer not allowed");
				}
			}
		}
	}

	//REFERER URL: Deny certain characters
	char c;	int i;
	for(i=0;i<Settings.Referrer.DenyCharacters.GetLength();i++){
		c = Settings.Referrer.DenyCharacters.GetAt(i);
		hack |= DenyChar(ReferrerURL,c,"'" + CString(c) + "' not allowed in Referrer URL");
	}

	//REFERER URL: Deny high bit characters (shellcode)
	if(Settings.Referrer.DenyHighBitShellcode){
		if(ScanHighBitShellCode(ReferrerURL)){
			hack = true;
			Alert("High bit characters (shellcode) not allowed in Referrer URL");
		}
	}

	//REFERER URL: Deny special whitespace characters
	if(Settings.Referrer.DenySpecialWhitespace){
		hack |= DenySpecialWhitespace(ReferrerURL, " not allowed in Referrer URL");
	}

	CStringHelper::Safe_MakeLower(ReferrerURL);

	//REFERER URL: Deny SQL injection
	if(Settings.Referrer.DenySQLInjection){
		int sqlcount(0);
		CStringList Matches;
		sqlcount += ScanSQLInjection(ReferrerURL,Settings.SQLKeyWords,Matches);
		if(sqlcount>0){
			if(sqlcount>1){
				hack = true;
				Alert("Possible SQL injection in referrer URL: " + CStringHelper::ToString(Matches,","));
			}else{
				logger.Append("WARNING: SQL keyword found in referrer URL (1 more will block request): " + CStringHelper::ToString(Matches,","));
			}
		}
	}

	//REFERER URL: Deny certain sequences (list should be lowercase)
	if(Settings.Referrer.DenySequences.Enabled){
		hack |= DenySequences(ReferrerURL,Settings.Referrer.DenySequences.List,"'","' not allowed in Referrer URL");
	}

	return hack;
}

CString CHTTPFirewall::GetDirectResponse(bool SendMessageBody)
{
	CString buf("HTTP/1.1 ");
	if(Settings.HTTPResponse.UseStatus){
		buf += Settings.HTTPResponse.Status;
	}else{
		buf += "404 Object Not Found";
	}
	if(!Settings.ResponseMonitor.RemoveServerHeader){
		buf += "\r\nServer: ";
		if(Settings.ResponseMonitor.ChangeServerHeader){
			buf += Settings.ResponseMonitor.ServerHeader;
		}else{
			buf += Settings.Admin.ServerHeader;
		}
	}
	_tzset();
	CString d = CTime::GetCurrentTime().FormatGmt("%a, %d %b %Y %H:%M:%S GMT");
	buf += "\r\nDate: " + d;
	buf += "\r\nContent-Type: text/html; charset=windows-1252";
	buf += "\r\nContent-Length: "; buf += CStringHelper::Safe_IntToAscii(Settings.HTTPResponse.Contents.GetLength());
	buf += "\r\nPragma: no-cache";
	buf += "\r\nCache-control: no-cache";
	buf += "\r\nExpires: " + d;
	buf += CRLFCRLF;
	if(SendMessageBody){ //see e-mail: special feature request of Sanford Whiteman if HEAD request -> no message body
		buf += Settings.HTTPResponse.Contents;
	}
	return buf;
}

void CHTTPFirewall::HackResponseBlockOrMonitor(CString& IP)
{
	//block or monitor IP address
	if(Settings.HTTPResponse.MonitorIP){
		IP_HACK_Monitored.AddNotInCache(IP);
	}
	if(Settings.HTTPResponse.BlockIP.Enabled){
		IP_HACK_Blocked.Add(IP);
	}
}
CString CHTTPFirewall::GetAdmin(CString Message)
{
	CString ret = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\r\n<html>\r\n<head>\r\n<title>" + Settings.Admin.ServerHeader + "</title>\r\n<META name=\"ROBOTS\" content=\"NOINDEX, NOFOLLOW\">\r\n";
	
	//toggle display
	ret += " <script language=\"javascript\" type=\"text/javascript\">\r\n";
    ret += "    function ToggleDisplay(id) { ";
    ret += "        var divstyle = new String(); ";
    ret += "        divstyle = document.getElementById(id).style.display; ";
    ret += "        if (divstyle == \"block\" || divstyle == \"\") { ";
    ret += "            document.getElementById(id).style.display = \"none\"; ";
    ret += "        } ";
    ret += "        else { ";
    ret += "            document.getElementById(id).style.display = \"block\"; ";
    ret += "        } ";
    ret += "    } ";
	ret += " </script>\r\n";

	ret += "</head>\r\n";

	//Body
	ret += "<body text=\"#000000\" vlink=\"#FF3300\" link=\"#FF3300\" bgcolor=\"#ffffff\"><font face=\"Verdana,Arial,Helvetica\" size=\"2\">\r\n";

	ret += "<h2>" + Settings.Admin.ServerHeader + "</h2>\r\n";

	//Admin
	ret += "<p><a href=\"?\">Home</a> - <a href=\"?Log\">Log</a> - <a href=\"?Settings\">Settings</a> - <a href=\"?Robots\">Robots</a> - <a href=\"?ReadMe\">Readme</a></p>\r\n";
	ret += "<p><strong>" + Message + "</strong></p>\r\n";


	//Stats
	ret += "<h3>Statistics</h3>\r\n";
	ret += "<noscript><p><strong><font color=\"red\">This functionality requires JavaScript to be enabled!</font></strong></p></noscript>\r\n";

	//Stringcaches
	ret += GetStatistics("Alerts",Settings.Statistics_Alert);
	ret += GetStatistics("IP_Authentication",IP_Authentication);
	ret += GetStatistics("IP_RequestsLimit",IP_RequestsLimit);
	ret += GetStatistics("IP_HACK_Monitored",IP_HACK_Monitored);
	ret += GetStatistics("IP_HACK_Blocked",IP_HACK_Blocked);
	ret += GetStatistics("IP_RESPONSE_Client_Errors",IP_RESPONSE_Client_Errors);
	ret += GetStatistics("IP_RESPONSE_Server_Errors",IP_RESPONSE_Server_Errors);
	ret += GetStatistics("IP_UserAgent",IP_UserAgent);
	ret += GetStatistics("Extension_RequestsLimit",Extension_RequestsLimit);
	ret += GetStatisticsAdditionalCaches();

	//Parsed IP ranges
	ret += "<h3>IP Ranges</h3>\r\n";
	if(Settings.Connection.MonitorAddresses.Enabled)
		ret += GetStatistics("MonitorIPRangeList",Settings.MonitorIPRangeList);
	if(Settings.Connection.DenyAddresses.Enabled)
		ret += GetStatistics("DenyIPRangeList",Settings.DenyIPRangeList);
	if(Settings.Engine.ExcludedIPAddresses.Enabled)
		ret += GetStatistics("ExcludedIPRangeList",Settings.ExcludedIPRangeList);

	//Regex patterns
	ret += "<h3>Regex Patterns</h3>";
	CString k;
	CAtlRegExp<>* regex;
	POSITION pos = Settings.regex_cache.GetStartPosition();
	if(pos!=NULL){
		ret += "<table border=\"1\">\r\n";
		while(pos!=NULL){
			Settings.regex_cache.GetNextAssoc(pos,k,(void*&)regex);
			ret += "<tr><td>" + CHTML::HTMLEncode(k) + "</td><td>";
			(regex==NULL) ? ret += "<font color=\"red\">ERROR</font>": ret+= "<font color=\"green\">OK</font>";
			ret += "</td></tr>\r\n";
		}
		ret += "</table>\r\n";
	}

	//show server time (local/GMT)
	ret += "<h3>Server</h3>\r\n<p>";
	ret += CTime::GetCurrentTime().Format("%Y-%m-%d %H:%M:%S");
	ret += " (local server time)\r\n<br>";
	ret += CTime::GetCurrentTime().FormatGmt("%Y-%m-%d %H:%M:%S");
	ret += " (GMT/UTC time)</p>\r\n";

	ret += CollapsibleTable("Admin",CStringHelper::ToString(Settings.Admin.FromIP,"|||"),Settings.Admin.FromIP.GetCount());
	ret += "<p>This page is only accessible from these IP addresses.</p>\r\n";
	
	//end
	ret += "</font>\r\n</body>\r\n</html>";

	return ret;	
}

CString CHTTPFirewall::GetStatistics(LPCTSTR Title, CStringCache& cache)
{
	return CollapsibleTable(Title,cache.ToString("|||"),cache.Count());
}

CString CHTTPFirewall::GetStatistics(LPCTSTR Title, CIPRangeList& list)
{
	return CollapsibleTable(Title,list.ToString("|||"),list.Count());
}

CString CHTTPFirewall::GetStatisticsAdditionalCaches()
{
	return "";
}

CString CHTTPFirewall::CollapsibleTable(LPCTSTR Title, CString& Text, unsigned int count)
{
	//TODO: ADMIN - make collabsible table also sortable

	//title
	CString ret = "<p>";
	ret += Title;
	ret += " (" + CStringHelper::Safe_IntToAscii(count) + ")";
	if(count>0){
		ret += " <a href=\"javascript:ToggleDisplay(\'";
		ret += Title;
		ret += "\');\">+</a>";
	}
	ret += "</p>\r\n";
	//items
	ret += "<div id=\"";
	ret += Title;
	ret += "\" style=\"display:none;\">\r\n";
	ret += "<table border=\"1\">\r\n<tr><td>";
	CString contents = CHTML::HTMLEncode(Text);
	contents.Replace("|||","</td></tr><tr><td>");
	ret += contents;
	ret += "</td></tr>\r\n</table>\r\n</div>\r\n";

	return ret;
}

